#uses "unSystemIntegrity/unSystemIntegrity_alarmDescription.ctl"

/**@name LIBRARY: unSystemIntegrity_DIP.ctl

@author: Marcin Bes (BE/ICS)

Creation Date: 23/06/2017

version 1.0

External Functions:

Internal Functions:

Purpose:
	Library of the DIP component used by the systemIntegrity.

Usage: Public

PVSS manager usage: CTRL (WCCOActrl)

Constraints:
	. global variable: the following variables are global to the script
	. data point type needed: _UnSystemIntegrity
	. data point: DIP_systemIntegrity
	. PVSS version: 3.0
	. operating system: W2000, NT and Linux, but tested only under W2000 and Linux.
	. distributed system: yes.
*/

#uses "fwDIP/fwDIP.ctl"

const int c_unSystemIntegrity_defaultDIPCheckingDelay = 30;
const string UN_SYSTEM_INTEGRITY_DIP = "DIP";
const string UN_SYSTEM_INTEGRITY_DIP_check = "unSystemIntegrity_DIPCheck";

// global declaration
global dyn_string g_unSystemIntegrity_DIP_DSList; // list of the DIP-DS dp checked
global dyn_int g_unSystemIntegrity_DIP_DSThreadId; // list of the DIP-DS thread Id checking the counter
global int g_unSystemIntegrity_DIP_CheckingDelay;
global bool g_bDIP_NotStarted = true;

//@{

//------------------------------------------------------------------------------------------------------------------------
// DIP_systemIntegrityInfo
/** Return the list of systemAlarm pattern.

@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  CTRL

@return list of systemAlarm pattern
*/
dyn_string DIP_systemIntegrityInfo() {
    return makeDynString(DS_pattern, DS_Time_pattern);
}

//------------------------------------------------------------------------------------------------------------------------

// unSystemIntegrity_DIP_Initialize
/**
Purpose:
Get the list of DIP check by the systemIntegrity. The ones that are enabled.

@param dsResult: dyn_string, output, the list of enabled DIP that are checked

Usage: Public

PVSS manager usage: CTRL (WCCOActrl)

Constraints:
	. PVSS version: 3.0
	. operating system: W2000, NT and Linux, but tested only under W2000 and Linux.
	. distributed system: yes.
*/
unSystemIntegrity_DIP_Initialize(dyn_string &dsResult) {
    dyn_string dsList;
    int len, i;
    string dpToCheck;
    bool enabled;

    dpGet(UN_SYSTEM_INTEGRITY_DIP + UN_SYSTEMINTEGRITY_EXTENSION + ".config.data", dsList);
    unSystemIntegrity_DIP_DataCallback("", dsList);

    dsList = dpNames(c_unSystemAlarm_dpPattern + DS_pattern + "*", c_unSystemAlarm_dpType);
    len = dynlen(dsList);
    for (i = 1; i <= len; i++) {
        dpToCheck = dpSubStr(dsList[i], DPSUB_DP);
        dpToCheck = substr(dpToCheck, strlen(c_unSystemAlarm_dpPattern + DS_pattern), strlen(dpToCheck));
        if (dpExists(dpToCheck)) {
            if (dpTypeName(dpToCheck) == UNDIP_DPTYPE) {
                dpGet(dsList[i] + ".enabled", enabled);
                if (enabled) {
                    dynAppend(dsResult, dpToCheck);
                }
            }
        }
    }
}

//------------------------------------------------------------------------------------------------------------------------

// unSystemIntegrity_DIP_DataCallback
/**
Purpose:
handle any systemIntegrity command.

@param sDpe1: string, input, dpe
@param sConfigData: string, input, the config data

Usage: Public

PVSS manager usage: CTRL (WCCOActrl)

Constraints:
	. PVSS version: 3.0
	. operating system: W2000, NT and Linux, but tested only under W2000 and Linux.
	. distributed system: yes.
*/
unSystemIntegrity_DIP_DataCallback(string sDpe1, dyn_string dsConfigData) {
    dyn_string dsTemp = dsConfigData;

    while (dynlen(dsTemp) <= 2) {
        dynAppend(dsTemp, "0");
    }

    g_unSystemIntegrity_DIP_CheckingDelay = (int)dsTemp[1];
    if (g_unSystemIntegrity_DIP_CheckingDelay <= 0) {
        g_unSystemIntegrity_DIP_CheckingDelay = c_unSystemIntegrity_defaultDIPCheckingDelay;
    }

//DebugN(dsConfigData, g_unSystemIntegrity_DIP_CheckingDelay);
}

//------------------------------------------------------------------------------------------------------------------------

// unSystemIntegrity_DIP_HandleCommand
/**
Purpose:
handle any systemIntegrity command.

@param sDpe1: string, input, dpe
@param command: int, input, the systemIntegrity command
@param sDpe2: string, input, dpe
@param parameters: dyn_string, input, the list of parameters of the systemIntegrity command

Usage: Public

PVSS manager usage: CTRL (WCCOActrl)

Constraints:
	. PVSS version: 3.0
	. operating system: W2000, NT and Linux, but tested only under W2000 and Linux.
	. distributed system: yes.
*/
unSystemIntegrity_DIP_HandleCommand(string sDpe1, int command, string sDpe2, dyn_string parameters) {
    dyn_string exceptionInfo;
    int i, len = dynlen(parameters);

    switch (command) {
        case UN_SYSTEMINTEGRITY_ADD: // add the _UnSystemAlarm DP
            for (i = 1; i <= len; i++) {
                unSystemIntegrity_DIP_checking(parameters[i], true, true, exceptionInfo);
            }
            break;
        case UN_SYSTEMINTEGRITY_DELETE: // delete the _UnSystemAlarm DP
            for (i = 1; i <= len; i++) {
                unSystemIntegrity_DIP_checking(parameters[i], false, false, exceptionInfo);
                // remove dp from the alert list of applicationDP if it is in
                _unSystemIntegrity_modifyApplicationAlertList(c_unSystemAlarm_dpPattern + DS_pattern + parameters[i], false, exceptionInfo);
                // delete the _UnSystemAlarm dps
                if (dpExists(c_unSystemAlarm_dpPattern + DS_pattern + parameters[i])) {
                    dpDelete(c_unSystemAlarm_dpPattern + DS_pattern + parameters[i]);
                }

                // remove dp from the alert list of applicationDP if it is in
                _unSystemIntegrity_modifyApplicationAlertList(c_unSystemAlarm_dpPattern + DS_Time_pattern + parameters[i], false, exceptionInfo);
                // delete the _UnSystemAlarm dps
                if (dpExists(c_unSystemAlarm_dpPattern + DS_Time_pattern + parameters[i])) {
                    dpDelete(c_unSystemAlarm_dpPattern + DS_Time_pattern + parameters[i]);
                }
            }
            break;
        case UN_SYSTEMINTEGRITY_ENABLE: // enable the _UnSystemAlarm DP
            for (i = 1; i <= len; i++) {
                if (dpExists(c_unSystemAlarm_dpPattern + DS_pattern + parameters[i])) {
                    unSystemIntegrity_DIP_checking(parameters[i], false, true, exceptionInfo);
                }
            }
            break;
        case UN_SYSTEMINTEGRITY_DISABLE: // disable the _UnSystemAlarm DP
            for (i = 1; i <= len; i++) {
                if (dpExists(c_unSystemAlarm_dpPattern + DS_pattern + parameters[i])) {
                    unSystemIntegrity_DIP_checking(parameters[i], false, false, exceptionInfo);
                }
            }
            break;
        case UN_SYSTEMINTEGRITY_DIAGNOSTIC: // give the list of _UnSystemAlarm DP and the state
            dpSet(UN_SYSTEM_INTEGRITY_DIP + UN_SYSTEMINTEGRITY_EXTENSION + ".diagnostic.commandResult", command,
                  UN_SYSTEM_INTEGRITY_DIP + UN_SYSTEMINTEGRITY_EXTENSION + ".diagnostic.result", g_unSystemIntegrity_DIP_DSList);
            break;
        default:
            break;
    }

    if (dynlen(exceptionInfo) > 0) {
        if (isFunctionDefined("unMessageText_sendException")) {
            unMessageText_sendException("*", "*", "unSystemIntegrity_DIP_HandleCommand", "user", "*", exceptionInfo);
        }
// handle any error uin case the send message failed
        if (dynlen(exceptionInfo) > 0) {
            DebugN(getCurrentTime(), exceptionInfo);
        }
    }

//DebugN(sDpe1, command, parameters, g_unSystemIntegrity_DIP_DSList);
}

//------------------------------------------------------------------------------------------------------------------------

// unSystemIntegrity_DIP_checking
/**
Purpose:
This function register/de-register the callback funtion of DIP manager. This function can also create the _unSystemAlarm_DIP dp
with the alarm config but it cannot delete it.

	@param sDp: string, input, data point name
	@param bCreate: bool, input, true create the dp and the alarm config if it does not exist, enable it
	@param bRegister: bool, input, true do a dpConnect, false do a dpDisconnect
	@param exceptionInfo: dyn_string, output, exception are returned here

Usage: Internal

PVSS manager usage: CTRL (WCCOActrl)

Constraints:
	. PVSS version: 3.0
	. operating system: W2000, NT and Linux, but tested only under W2000 and Linux.
	. distributed system: yes.
*/

unSystemIntegrity_DIP_checking(string sDp, bool bCreate, bool bRegister, dyn_string &exceptionInfo) {
    string sDpToChek, dp, dpTime, dpError, dpPlc, plcHostname, description;
    int res, thId, pos, manNum = 0, len, i;
    bool bError = false;
    dyn_string plc_plc_list;

    string alertPanel, alertHelp;
    bool configExists, isActive;
    int alertConfigType;
    dyn_string alertTexts, alertClasses, summaryDpeList, alertPanelParameters;
    dyn_float alertLimits;

// remove the system name
    res = strpos(sDp, ":");
    if (res >= 0) {
        dp = substr(sDp, res + 1, strlen(sDp));
    } else {
        dp = sDp;
    }
    sDpToChek = dp;
    dpTime = c_unSystemAlarm_dpPattern + DS_Time_pattern + dp;
    dp = c_unSystemAlarm_dpPattern + DS_pattern + dp;
// DebugN("unSystemIntegrity_DIP_checking", dp, bCreate, bRegister, sDpToChek, dpTime);

// in case the Dp is not of type _UnPlc do nothing. Should not happen, just for security.
    if (dpTypeName(sDpToChek) != UNDIP_DPTYPE) {
        return;
    }

    dyn_int diNumber;
    dyn_string dsConfigFileDIPConfig;

    if(_fwDIP_getAllDIPConfigFromConfigFile(diNumber, dsConfigFileDIPConfig) != 0) {
        fwException_raise(exceptionInfo,"ERROR","Cannot retrieve DIP config info from config file.","");
    } else {
        string sDIPConfigDP = "DIPConfig_" + unGenericDpFunctions_getAlias(sDpToChek);
        int iResult = dynContains(dsConfigFileDIPConfig, sDIPConfigDP);
        if(iResult > 0) {
            manNum = diNumber[iResult]; // read the DIP manager number
        }
    }

    if (bCreate) {
// get the plc hostname and the driver number
        plcHostname = unGenericDpFunctions_getAlias(sDpToChek);
// create the sDp and its alarm config if it is not existing
        if (!dpExists(dp)) {
            description = getCatStr("unSystemIntegrity", "PLC_DS_DESCRIPTION") + plcHostname + " -> DS DIP manager " + manNum;
            unSystemIntegrity_createSystemAlarm(sDpToChek, DS_pattern, description, exceptionInfo);
        }
    }

    // handle alarm description connection/disconnection
    shared_ptr<unSystemIntegrity_UpdateAlarmDescription> updAlDesc = new unSystemIntegrity_UpdateAlarmDescription();
    if (bRegister) {
        // register alarm description
        unSystemIntegrity_registerAlarmDescription(dp + ".alarm", updAlDesc, exceptionInfo);
        classConnect("unSystemIntegrity_DIP_alarmDescriptionCB", updAlDesc, unSystemIntegrity_UpdateAlarmDescription::updateAlarmDescription);

// connect to the callback function if not already done
        pos = _unSystemIntegrity_isInList(g_unSystemIntegrity_DIP_DSList, dp);
        if (pos <= 0) {
            // add it in the list, because one of the callback function needs it.
            pos = _unSystemIntegrity_setList(g_unSystemIntegrity_DIP_DSList, dp, true);

// start the thread for checking the counter
            thId = startThread(UN_SYSTEM_INTEGRITY_DIP_check, sDpToChek, manNum);
            g_unSystemIntegrity_DIP_DSThreadId[pos] = thId;
            if (thId < 0) {
                bError = true;
            }

            if (bError) {

                pos = _unSystemIntegrity_isInList(g_unSystemIntegrity_DIP_DSList, dp);
                if (pos > 0) {
                    // kill the threads if they were started
                    res = stopThread(g_unSystemIntegrity_DIP_DSThreadId[pos]);
// and then remove it from the list
                    pos = _unSystemIntegrity_setList(g_unSystemIntegrity_DIP_DSList, dp, false);
//DebugN("rmv", res);
                    dynRemove(g_unSystemIntegrity_DIP_DSThreadId, pos);
                }

                fwException_raise(exceptionInfo, "ERROR",
                                  "unSystemIntegrity_DIP_checking():" + getCatStr("unSystemIntegrity", "CANNOT_DP_CONNECT") + dp, "");
            } else {
//DebugN("Enable", g_unSystemIntegrity_DIP_DSThreadId, "end");
                // set the enable to true and activate the alarm if it is not activated.
                dpSet(dp + ".enabled", true);
                unAlarmConfig_mask(dp + ".alarm", false, exceptionInfo);
                // get all the PLC-PLC dp, and enable them
                plc_plc_list = dpNames(c_unSystemAlarm_dpPattern + PLC_PLC_pattern + sDpToChek + "*", c_unSystemAlarm_dpType);
                len = dynlen(plc_plc_list);
                for (i = 1; i <= len; i++) {
                    dpSet(plc_plc_list[i] + ".enabled", true);
                    unAlarmConfig_mask(plc_plc_list[i] + ".alarm", false, exceptionInfo);
                }

                // get all the FESystemAlarm dp, and enable them, check the Ok state and set the correct value
                plc_plc_list = dpNames(c_unSystemAlarm_dpPattern + FE_pattern + sDpToChek + "*", c_unSystemAlarm_dpType);
                len = dynlen(plc_plc_list);
                for (i = 1; i <= len; i++) {
                    dpSet(plc_plc_list[i] + ".enabled", true);
                    unAlarmConfig_mask(plc_plc_list[i] + ".alarm", false, exceptionInfo);
                }
            }
        }
    } else {
        // unregister alarm description
        classDisconnect("unSystemIntegrity_DIP_alarmDescriptionCB", updAlDesc, unSystemIntegrity_UpdateAlarmDescription::updateAlarmDescription);

// disconnect the callback function
//DebugN("Disable", g_unSystemIntegrity_DIP_DSThreadId, "end");
        pos = _unSystemIntegrity_isInList(g_unSystemIntegrity_DIP_DSList, dp);
        if (pos > 0) {
// kill the thread
            res = stopThread(g_unSystemIntegrity_DIP_DSThreadId[pos]);
            if (res < 0) {
                bError = true;
            }

            if (bError) {
                fwException_raise(exceptionInfo, "ERROR",
                                  "unSystemIntegrity_DIP_checking():" + getCatStr("unSystemIntegrity", "CANNOT_DP_DISCONNECT") + dp, "");
            } else {
                // remove from list
                pos = _unSystemIntegrity_setList(g_unSystemIntegrity_DIP_DSList, dp, false);
//DebugN("rmv", res);
                dynRemove(g_unSystemIntegrity_DIP_DSThreadId, pos);
                // set the enable to false and de-activate the alarm and acknowledge it if necessary
                dpSet(dp + ".enabled", false, dp + ".alarm", c_unSystemIntegrity_no_alarm_value);
                unAlarmConfig_mask(dp + ".alarm", true, exceptionInfo);
                // get all the PLC-PLC dp, and disable them
                plc_plc_list = dpNames(c_unSystemAlarm_dpPattern + PLC_PLC_pattern + sDpToChek + "*", c_unSystemAlarm_dpType);
                len = dynlen(plc_plc_list);
                for (i = 1; i <= len; i++) {
// do not reset the alarm value because this DPE should be connected to an address config
                    dpSet(plc_plc_list[i] + ".enabled", false);
                    unAlarmConfig_mask(plc_plc_list[i] + ".alarm", true, exceptionInfo);
                }

                // get all the FESystemAlarm dp, and disable them, check the Ok state and set the correct value
                plc_plc_list = dpNames(c_unSystemAlarm_dpPattern + FE_pattern + sDpToChek + "*", c_unSystemAlarm_dpType);
                len = dynlen(plc_plc_list);
                for (i = 1; i <= len; i++) {
// do not reset the alarm value because this DPE should be connected to an address config
                    dpSet(plc_plc_list[i] + ".enabled", false);
                    unAlarmConfig_mask(plc_plc_list[i] + ".alarm", true, exceptionInfo);
                }
            }
        }
    }
}

//------------------------------------------------------------------------------------------------------------------------

// unSystemIntegrity_DIPCheck
/**
Purpose:
This function checks if the counter of the DIP dp is modified periodically. If not it generates and alarm.

	@param dp: string, input, data point name

Usage: Internal

PVSS manager usage: CTRL (WCCOActrl)

Constraints:
	. PVSS version: 3.0
	. operating system: W2000, NT and Linux, but tested only under W2000 and Linux.
	. distributed system: yes.
*/
unSystemIntegrity_DIPCheck(string dp, int iDrvNum) {
    int waitingTime, oldAlarmValue = -1;
    long oldValue = -1, newValue, alarmValue;
    bool bEnabled;
    string sModPlc;
    int res, iPlcNumber;

    while (true) {
        // if 0 set it to g_unSystemIntegrity_DIP_CheckingDelay
        waitingTime = g_unSystemIntegrity_DIP_CheckingDelay;
        if (waitingTime <= 0) {
            waitingTime = c_unSystemIntegrity_defaultDIPCheckingDelay;
        }
        // wait
        delay(waitingTime);

		if (!dpExists(dp)) {
			DebugTN(__FUNCTION__,"DP does not exist anymore: "+dp,"Likely frontend dropped - abandoning the moniroring");
			return;
		}
        // get the value of the counter and the timeout PLCChecking
        dpGet(dp + ".communication.counter:_original.._stime", newValue);
        if (newValue == oldValue) { // if same value, check if same time
            // counter was not modified and same timestamp, set an alarm
            alarmValue = c_unSystemIntegrity_alarm_value_level1;
        } else
            // counter was modified, reset the alarm
        {
            alarmValue = c_unSystemIntegrity_no_alarm_value;
        }

        if (oldAlarmValue != alarmValue) {
            dpSet(c_unSystemAlarm_dpPattern + DS_pattern + dp + ".alarm", alarmValue);
        }

//DebugN("Check counter", dp, oldValue, newValue, alarmValue);
        oldValue = newValue;
        oldAlarmValue = alarmValue;
    }
}

//------------------------------------------------------------------------------------------------------------------------

// unSystemIntegrity_DIP_alarmDescriptionCB
/**
* Callback function to set the DIP alarm description in the alertSet when
* the updateAlarmDescription class event has been triggered
*
* @param timeStamp         input, time, alert timestamp
* @param aTimeStamp        input, atime, alert atimestamp
* @param dpeAlert          input, string, alert datapoint element
* @param alarmValue        input, int, alarm value number
*/

public void unSystemIntegrity_DIP_alarmDescriptionCB(time timeStamp, atime aTimeStamp, string dpeAlert, int alarmValue) {
  string alarmDescription = getCatStr("DIP_systemIntegrity",DS_pattern+alarmValue);
  unSystemIntegrity_setAlarmDescription(alarmDescription, timeStamp, aTimeStamp, dpeAlert);
}

//@}

